@@ -2,25 +2,31 @@ name = "knockpy"
 all = [
 	"constants",
     "dgp",
+    "ggm",
     "knockoff_stats",
     "knockoff_filter",
     "knockoffs",
     "mac",
     "metro",
+    "mlr",
     "mrc",
     "smatrix",
     "utilities",
 ]
 
-__version__ = "1.0.4"
+__version__ = "1.3.0"
 
 from . import constants
 from . import dgp
+from . import ggm
 from . import knockoffs
 from . import knockoff_filter
 from . import knockoff_stats
 from . import mac
 from . import metro
+from . import mlr
 from . import mrc
 from . import smatrix
 from . import utilities
+
+from .knockoff_filter import KnockoffFilter
