@@ -5,6 +5,7 @@ from . import utilities
 from . import mrc
 from . import knockoffs
 from . import metro
+from . import mlr
 from . import knockoff_stats as kstats
 
 
@@ -19,25 +20,33 @@ class KnockoffFilter:
     ----------
     fstat : str or knockpy.knockoff_stats.FeatureStatistic
         The feature statistic to use in the knockoff filter.
+        Defaults to a lasso statistic.
         This may also be a string identifier, including:
-        - 'lasso' or 'lcd': cross-validated lasso coefficients differences 
-        - 'lsm': signed maximum of the lasso path statistic as 
-            in Barber and Candes 2015
+
+        - 'lasso' or 'lcd': cross-validated lasso coefficients differences
+        - 'mlr': the masked likelihood ratio (MLR) statistic, which has provable optimality properties for knockoffs
+        - 'lsm': signed maximum of the lasso path statistic as in Barber and Candes 2015
         - 'dlasso': Cross-validated debiased lasso coefficients
         - 'ridge': Cross validated ridge coefficients
         - 'ols': Ordinary least squares coefficients
         - 'margcorr': marginal correlations between features and response
         - 'deeppink': The deepPINK statistic as in Lu et al. 2018
+        - 'mlrspline' or 'spline': The MLR statistic applied to regression spline basis functions.
         - 'randomforest': A random forest with swap importances
+
+        Note that when using FX knockoffs, the `lcd` statistic does not use
+        cross-valdilidation and sets ``alphas = sigma2 * sqrt(log(p) / n).``
     ksampler : str or knockpy.knockoffs.KnockoffSampler
         The knockoff sampler to use in the knockoff filter.
         This may also be a string identifier, including:
+
         - 'gaussian': Gaussian Model-X knockoffs
-        - 'fx': Fixed-X knockoffs
-        - 'metro': Generic metropolized knockoff sampler.
-        - 'artk': t-tailed Markov chain
-        - 'blockt': Blocks of t-distributed 
+        - 'fx': Fixed-X knockoffs.
+        - 'metro': Generic metropolized knockoff sampler
+        - 'artk': t-tailed Markov chain.
+        - 'blockt': Blocks of t-distributed
         - 'gibbs_grid': Discrete gibbs grid
+
         An alternative to specifying the ksampler is to simply pass
         in a knockoff matrix during the ``forward`` call.
     fstat_kwargs : dict
@@ -116,6 +125,24 @@ class KnockoffFilter:
         Initialize the class.
         """
 
+        ### Preprocessing for knockoffs
+        self.knockoff_kwargs = knockoff_kwargs.copy()
+        if isinstance(ksampler, str):
+            self.knockoff_type = ksampler.lower()
+            self.ksampler = None
+        if isinstance(ksampler, knockoffs.KnockoffSampler):
+            self.knockoff_type = None
+            self.ksampler = ksampler
+        if isinstance(ksampler, knockoffs.FXSampler):
+            self._mx = False
+        elif self.knockoff_type == "fx":
+            self._mx = False
+        else:
+            self._mx = True
+
+        # Initialize
+        self.S = None
+
         ### Parse feature statistic
         self.fstat_kwargs = fstat_kwargs.copy()
         if isinstance(fstat, str):
@@ -143,28 +170,17 @@ class KnockoffFilter:
             fstat = kstats.RandomForestStatistic()
         elif fstat == "deeppink":
             fstat = kstats.DeepPinkStatistic()
+        elif fstat == 'mlr':
+            if self._mx:
+                fstat = mlr.MLR_Spikeslab()
+            else:
+                fstat = mlr.MLR_FX_Spikeslab()
+        elif fstat in ['mlr_spline', 'spline']:
+            fstat = mlr.MLR_Spikeslab_Splines()
         else:
             raise ValueError(f"Unrecognized fstat {fstat}")
         self.fstat = fstat
 
-        ### Preprocessing for knockoffs
-        self.knockoff_kwargs = knockoff_kwargs.copy()
-        if isinstance(ksampler, str):
-            self.knockoff_type = ksampler.lower()
-            self.ksampler = None
-        if isinstance(ksampler, knockoffs.KnockoffSampler):
-            self.knockoff_type = None
-            self.ksampler = ksampler
-        if isinstance(ksampler, knockoffs.FXSampler):
-            self._mx = False
-        elif self.knockoff_type == "fx":
-            self._mx = False
-        else:
-            self._mx = True
-
-        # Initialize
-        self.S = None
-
     def sample_knockoffs(self):
         """
         Samples knockoffs during ``forward``.
@@ -224,17 +240,10 @@ class KnockoffFilter:
                 ],
                 axis=1,
             )
-            # Handle errors where Ginv is exactly low rank
-            try:
-                self.Ginv = utilities.chol2inv(self.G)
-            except np.linalg.LinAlgError:
-                warnings.warn("The feature-knockoff covariance matrix is low rank.")
-                self.Ginv = None
+            self.Ginv = None
         else:
-            self.G, self.Ginv = utilities.estimate_covariance(
-                np.concatenate([self.X, self.Xk], axis=1)
-            )
-
+            self.G = None
+            self.Ginv = None
         return self.Xk
 
     def make_selections(self, W, fdr):
@@ -301,11 +310,11 @@ class KnockoffFilter:
             with rank ``num_factors`` to speed up computation.
         recycle_up_to : int or float
             Three options:
+
                 - if ``None``, does nothing.
-                - if an integer > 1, uses the first "recycle_up_to"
-                rows of X as the the first ``recycle_up_to`` rows of knockoffs
-                - if a float between 0 and 1 (inclusive), interpreted
-                as the proportion of rows to recycle. 
+                - if an integer > 1, uses the first "recycle_up_to" rows of X as the the first ``recycle_up_to`` rows of knockoffs
+                - if a float between 0 and 1 (inclusive), interpreted as the proportion of rows to recycle. 
+
             For more on recycling, see https://arxiv.org/abs/1602.03574
         """
 
@@ -331,6 +340,9 @@ class KnockoffFilter:
         # Save objects
         self.X = X
         self.Xk = Xk
+        # Center if we are going to fit FX knockoffs
+        if self.Xk is None and not self._mx:
+            self.X = self.X - self.X.mean(axis=0)
         self.y = y
         self.mu = mu
         self.Sigma = Sigma
@@ -359,9 +371,29 @@ class KnockoffFilter:
         if self.Xk is None:
             self.Xk = self.sample_knockoffs()
 
+        # When using FX knockoffs and lcd statistic,
+        # we cannot use cross-validation. Thus, we pick
+        # a default regularization parameter
+        if not self._mx and isinstance(
+            self.fstat, kstats.LassoStatistic
+        ):
+            if self.fstat_kwargs.get('zstat', 'coef') != 'lars_path':
+                hatsigma2 = kstats.compute_residual_variance(
+                    self.X, self.Xk, self.y
+                )
+                self.fstat_kwargs['alphas'] = hatsigma2 * np.sqrt(np.log(p)/n)
+
+
         # As an edge case, pass Ginv to debiased lasso
         if "debias" in self.fstat_kwargs:
             if self.fstat_kwargs["debias"]:
+                if self.Ginv is None:
+                    if self.G is not None:
+                        self.Ginv = np.linalg.inv(self.G)
+                    else:
+                        self.G, self.Ginv = utilities.estimate_covariance(
+                           np.concatenate([self.X, self.Xk], axis=1)
+                        )
                 self.fstat_kwargs["Ginv"] = self.Ginv
 
         # Feature statistics
@@ -369,11 +401,62 @@ class KnockoffFilter:
             X=self.X, Xk=self.Xk, y=self.y, groups=self.groups, **self.fstat_kwargs
         )
         # Inherit some attributes
+        self.fdr = fdr
         self.Z = self.fstat.Z
         self.W = self.fstat.W
         self.score = self.fstat.score
         self.score_type = self.fstat.score_type
-        self.rejections = self.make_selections(self.W, fdr)
+        self.rejections = self.make_selections(self.W, self.fdr)
 
         # Return
         return self.rejections
+
+    def seqstep_plot(
+        self,
+        max_rank=None,
+    ):
+        """
+        Visualizes the knockoff filter.
+
+        Parameters
+        ----------
+        max_rank : int
+            Only plot the top ``max_rank`` feature statistics.
+            Defaults to plotting all feature statistics.
+        """
+        # Import matplotlib
+        try:
+            import matplotlib.pyplot as plt
+        except ImportError as err:
+            raise ImportError(
+                f"matplotlib is required for plotting, but importing it raised {err}."
+            )
+        if max_rank is None:
+            max_rank = self.X.shape[1]
+
+        # sort W statistics
+        inds = np.argsort(-1*np.abs(self.W))
+        sortW = self.W[inds][0:max_rank]
+        xvals = np.arange(max_rank)
+
+        # threshold
+        threshold_ind = np.sum(np.abs(self.W) >= self.threshold)
+
+        # Plot and show
+        plt.bar(xvals[sortW < 0], sortW[sortW < 0], color='red')
+        plt.bar(xvals[sortW >= 0], sortW[sortW >= 0], color='blue')
+        if threshold_ind >= max_rank:
+            print("Not plotting threshold since threshold <= max_rank.")
+        else:
+            plt.axvline(
+                threshold_ind, 
+                color='black', 
+                linestyle='dashed', 
+                label=f'Threshold\n(q={self.fdr})'
+            )
+            plt.legend()
+        plt.xlabel("Rank")
+        plt.ylabel("Feature statistic")
+        plt.show()
+
+
