@@ -4,8 +4,6 @@ import numpy as np
 # Model-fitters
 import sklearn
 from sklearn import linear_model, model_selection, ensemble
-from group_lasso import GroupLasso, LogisticGroupLasso
-from pyglmnet import GLMCV
 from . import utilities
 from .constants import DEFAULT_REG_VALS
 
@@ -16,21 +14,6 @@ def calc_mse(model, X, y):
     resids = (preds - y) / y.std()
     return np.sum((resids) ** 2)
 
-
-def use_reg_lasso(groups):
-    """ Parses whether or not to use group lasso """
-    # See if we are using regular lasso...
-    if groups is not None:
-        p = groups.shape[0]
-        m = np.unique(groups).shape[0]
-        if p == m:
-            return True
-        else:
-            return False
-    else:
-        return True
-
-
 def parse_y_dist(y):
     """ Checks if y is binary; else assumes it is continuous """
     n = y.shape[0]
@@ -39,7 +22,7 @@ def parse_y_dist(y):
     elif np.unique(y).shape[0] == n:
         return "gaussian"
     else:
-        warnings.warn("Treating y data as continuous even though it may be discrete.")
+        #warnings.warn("Treating y data as continuous even though it may be discrete.")
         return "gaussian"
 
 
@@ -91,7 +74,9 @@ def combine_Z_stats(Z, groups=None, antisym="cd", group_agg="sum"):
             f"Unexpected shape {Z.shape} for Z statistics (expected ({2*p},))"
         )
     if groups is None:
-        groups = np.arange(1, p + 1, 1)
+        groups = np.arange(1, p + 1)
+    else:
+        groups = utilities.preprocess_groups(groups)
 
     antisym = str(antisym).lower()
     # Absolute coefficient differences
@@ -126,6 +111,36 @@ def combine_Z_stats(Z, groups=None, antisym="cd", group_agg="sum"):
     # Return
     return W_group
 
+def compute_residual_variance(X, Xk, y):
+    """
+    Estimates sigma2 using residual variance of y
+    given [X, Xk]. Uses a penalized linear model
+    if n < 2p + 10 or p >= 1500 (for efficiency).
+
+    Returns
+    -------
+    sigma2 : float
+        Estimated residual variance
+    """
+    # Use (penalized) linear model to predict y
+    n, p = X.shape
+    if n - 2 * p > 10 and p < 1500:
+        model = linear_model.LinearRegression()
+    else:
+        model = linear_model.Lasso(
+            alpha=0.1 * np.sqrt(np.log(p) / n),
+            max_iter=100
+        )
+    # Concatenating features shows no preference between X, Xk
+    features = np.concatenate([X, Xk], axis=1)
+    perm_inds, _ = utilities.random_permutation_inds(2 * p)
+    features = features[:, perm_inds]
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        model.fit(features, y)
+    resid = np.power(model.predict(features) - y, 2).sum()
+    return resid / max(10, n - np.sum(model.coef_ != 0))
+
 
 # ------------------------------ Lasso Stuff ---------------------------------------#
 def calc_lars_path(X, Xk, y, groups=None, **kwargs):
@@ -186,7 +201,7 @@ def calc_lars_path(X, Xk, y, groups=None, **kwargs):
     return Z[rev_inds]
 
 
-def fit_lasso(X, Xk, y, y_dist=None, use_lars=False, **kwargs):
+def fit_lasso(X, Xk, y, y_dist=None, alphas=None, use_lars=False, **kwargs):
     """
     Fits cross-validated lasso on [X, Xk] and y.
     
@@ -200,6 +215,9 @@ def fit_lasso(X, Xk, y, y_dist=None, use_lars=False, **kwargs):
         ``(n,)``-shaped response vector
     y_dist : str
         One of "binomial" or "gaussian"
+    alphas : float
+        The regularization parameter(s) to try. Selects one
+        alpha using cross-validation by default.
     use_lars : bool
         If True, uses a LARS-based solver for Gaussian data.
         If False, uses a gradient based solver (default).
@@ -213,7 +231,7 @@ def fit_lasso(X, Xk, y, y_dist=None, use_lars=False, **kwargs):
 
     Returns
     -------
-    gl : sklearn.linear_model.LassoCV/LassoLarsCV/LogisticRegressionCV
+    gl : sklearn Lasso/LassoCV/LassoLarsCV/LogisticRegressionCV object
         The sklearn model fit through cross-validation.
     inds : np.ndarray
         ``(2p,)``-dimensional array of indices representing the random
@@ -228,6 +246,10 @@ def fit_lasso(X, Xk, y, y_dist=None, use_lars=False, **kwargs):
     max_iter = kwargs.pop("max_iter", 500)
     tol = kwargs.pop("tol", 1e-3)
     cv = kwargs.pop("cv", 5)
+    if alphas is None:
+        alphas = DEFAULT_REG_VALS
+    if isinstance(alphas, float) or isinstance(alphas, int):
+        alphas = [alphas]
     if y_dist is None:
         y_dist = parse_y_dist(y)
 
@@ -240,35 +262,42 @@ def fit_lasso(X, Xk, y, y_dist=None, use_lars=False, **kwargs):
     features = features[:, inds]
 
     # Fit lasso
-    warnings.filterwarnings("ignore")
-    if y_dist == "gaussian":
-        if not use_lars:
-            gl = linear_model.LassoCV(
-                alphas=DEFAULT_REG_VALS,
-                cv=cv,
-                verbose=False,
+    with warnings.catch_warnings():
+        warnings.simplefilter("ignore")
+        if y_dist == "gaussian":
+            if not use_lars and len(alphas) == 1:
+                gl = linear_model.Lasso(
+                    alpha=alphas[0],
+                    max_iter=max_iter,
+                    tol=tol,
+                    **kwargs
+                ).fit(features, y)
+            elif not use_lars:
+                gl = linear_model.LassoCV(
+                    alphas=alphas,
+                    cv=cv,
+                    verbose=False,
+                    max_iter=max_iter,
+                    tol=tol,
+                    **kwargs,
+                ).fit(features, y)
+            elif use_lars:
+                gl = linear_model.LassoLarsCV(
+                    cv=cv, verbose=False, max_iter=max_iter, **kwargs,
+                ).fit(features, y)
+        elif y_dist == "binomial":
+            gl = linear_model.LogisticRegressionCV(
+                Cs=1 / np.array(alphas),
+                penalty="l1",
                 max_iter=max_iter,
                 tol=tol,
+                cv=cv,
+                verbose=False,
+                solver="liblinear",
                 **kwargs,
             ).fit(features, y)
-        elif use_lars:
-            gl = linear_model.LassoLarsCV(
-                cv=cv, verbose=False, max_iter=max_iter, **kwargs,
-            ).fit(features, y)
-    elif y_dist == "binomial":
-        gl = linear_model.LogisticRegressionCV(
-            Cs=1 / DEFAULT_REG_VALS,
-            penalty="l1",
-            max_iter=max_iter,
-            tol=tol,
-            cv=cv,
-            verbose=False,
-            solver="liblinear",
-            **kwargs,
-        ).fit(features, y)
-    else:
-        raise ValueError(f"y_dist must be one of gaussian, binomial, not {y_dist}")
-    warnings.resetwarnings()
+        else:
+            raise ValueError(f"y_dist must be one of gaussian, binomial, not {y_dist}")
 
     return gl, inds, rev_inds
 
@@ -317,180 +346,24 @@ def fit_ridge(X, Xk, y, y_dist=None, **kwargs):
     features = features[:, inds]
 
     # Fit lasso
-    warnings.filterwarnings("ignore")
-    if y_dist == "gaussian":
-        ridge = linear_model.RidgeCV(
-            alphas=DEFAULT_REG_VALS,
-            store_cv_values=True,
-            scoring="neg_mean_squared_error",
-            **kwargs,
-        ).fit(features, y)
-    elif y_dist == "binomial":
-        ridge = linear_model.LogisticRegressionCV(
-            Cs=1 / DEFAULT_REG_VALS, penalty="l2", solver="liblinear", **kwargs,
-        ).fit(features, y)
-    else:
-        raise ValueError(f"y_dist must be one of gaussian, binomial, not {y_dist}")
-    warnings.resetwarnings()
+    with warnings.catch_warnings():
+        warnings.simplefilter('ignore')
+        if y_dist == "gaussian":
+            ridge = linear_model.RidgeCV(
+                alphas=DEFAULT_REG_VALS,
+                store_cv_values=True,
+                scoring="neg_mean_squared_error",
+                **kwargs,
+            ).fit(features, y)
+        elif y_dist == "binomial":
+            ridge = linear_model.LogisticRegressionCV(
+                Cs=1 / DEFAULT_REG_VALS, penalty="l2", solver="liblinear", **kwargs,
+            ).fit(features, y)
+        else:
+            raise ValueError(f"y_dist must be one of gaussian, binomial, not {y_dist}")
 
     return ridge, inds, rev_inds
 
-
-def fit_group_lasso(
-    X, Xk, y, groups, use_pyglm=True, y_dist=None, group_lasso=True, **kwargs,
-):
-    """
-    Fits cross-validated ridge on [X, Xk] and y.
-
-    Parameters
-    ----------
-    X : np.ndarray
-        the ``(n, p)``-shaped design matrix
-    Xk : np.ndarray
-        the ``(n, p)``-shaped matrix of knockoffs
-    y : np.ndarray
-        ``(n,)``-shaped response vector
-    groups : np.ndarray
-        For group knockoffs, a p-length array of integers from 1 to 
-        num_groups such that ``groups[j] == i`` indicates that variable `j`
-        is a member of group `i`. Defaults to None (regular knockoffs). 
-    use_pyglm : bool
-        When fitting the group lasso, use the pyglm package if True (default).
-        Else, use the group_lasso package.
-    y_dist : str
-        One of "binomial" or "gaussian"
-    group_lasso : bool
-        If True, use a true group lasso. Else just use the sklearn ungrouped
-        lasso.
-    **kwargs
-        kwargs for eventual (group) lasso model.
-
-    Notes
-    -----
-    To avoid FDR control violations, this randomly permutes
-    the features before fitting the group lasso.
-
-    Returns
-    -------
-    gl : pyglm/sklearn/group_lasso model
-        The model fit through cross-validation; one of many types.
-    inds : np.ndarray
-        ``(2p,)``-dimensional array of indices representing the random
-        permutation applied to the concatenation of [X, Xk] before fitting
-        ``gl.``
-    rev_inds : np.ndarray:
-        Indices which reverse the effect of ``inds.`` In particular, if
-        M is any ``(n, 2p)``-dimensional array, then ```M==M[:, inds][:, rev_inds]```
-    """
-
-    warnings.filterwarnings("ignore")
-
-    # Parse some kwargs/defaults
-    max_iter = kwargs.pop("max_iter", 100)
-    tol = kwargs.pop("tol", 1e-2)
-    cv = kwargs.pop("cv", 5)
-    learning_rate = kwargs.pop("learning_rate", 2)
-    if y_dist is None:
-        y_dist = parse_y_dist(y)
-
-    # Bind data
-    n = X.shape[0]
-    p = X.shape[1]
-
-    # By default, all variables are their own group
-    if groups is None:
-        groups = np.arange(1, p + 1, 1)
-    m = np.unique(groups).shape[0]
-
-    # If m == p, meaning each variable is their own group,
-    # just fit a regular lasso
-    if m == p or not group_lasso:
-        return fit_lasso(X, Xk, y, y_dist, **kwargs)
-
-    # Make sure variables and their knockoffs are in the same group
-    # This is necessary for antisymmetry
-    doubled_groups = np.concatenate([groups, groups], axis=0)
-
-    # Randomize coordinates to make sure everything is symmetric
-    inds, rev_inds = utilities.random_permutation_inds(2 * p)
-    features = np.concatenate([X, Xk], axis=1)
-    features = features[:, inds]
-    doubled_groups = doubled_groups[inds]
-
-    # Standardize - important for pyglmnet performance,
-    # highly detrimental for group_lasso performance
-    if use_pyglm:
-        features = (features - features.mean()) / features.std()
-        if y_dist == "gaussian":
-            y = (y - y.mean()) / y.std()
-
-    # Get regularization values for cross validation
-    reg_vals = kwargs.pop("reg_vals", [(x, x) for x in DEFAULT_REG_VALS])
-
-    # Fit pyglm model using warm starts
-    if use_pyglm:
-
-        l1_regs = [x[0] for x in reg_vals]
-
-        gl = GLMCV(
-            distr=y_dist,
-            tol=tol,
-            group=doubled_groups,
-            alpha=1.0,
-            learning_rate=learning_rate,
-            max_iter=max_iter,
-            reg_lambda=l1_regs,
-            cv=cv,
-            solver="cdfast",
-        )
-        gl.fit(features, y)
-
-        # Pull score, rename
-        best_score = -1 * calc_mse(gl, features, y)
-        best_gl = gl
-
-    # Fit model
-    if not use_pyglm:
-        best_gl = None
-        best_score = -1 * np.inf
-        for group_reg, l1_reg in reg_vals:
-
-            # Fit logistic/gaussian group lasso
-            if not use_pyglm:
-                if y_dist.lower() == "gaussian":
-                    gl = GroupLasso(
-                        groups=doubled_groups,
-                        tol=tol,
-                        group_reg=group_reg,
-                        l1_reg=l1_reg,
-                        **kwargs,
-                    )
-                elif y_dist.lower() == "binomial":
-                    gl = LogisticGroupLasso(
-                        groups=doubled_groups,
-                        tol=tol,
-                        group_reg=group_reg,
-                        l1_reg=l1_reg,
-                        **kwargs,
-                    )
-                else:
-                    raise ValueError(
-                        f"y_dist must be one of gaussian, binomial, not {y_dist}"
-                    )
-
-                gl.fit(features, y.reshape(n, 1))
-                score = -1 * calc_mse(gl, features, y.reshape(n, 1))
-
-            # Score, possibly select
-            if score > best_score:
-                best_score = score
-                best_gl = gl
-
-    warnings.resetwarnings()
-
-    return best_gl, inds, rev_inds
-
-
 class FeatureStatistic:
     """
     The base knockoff feature statistic class, which can wrap any
@@ -929,8 +802,6 @@ class LassoStatistic(FeatureStatistic):
         y,
         groups=None,
         zstat="coef",
-        use_pyglm=True,
-        group_lasso=False,
         antisym="cd",
         group_agg="avg",
         cv_score=False,
@@ -960,14 +831,6 @@ class LassoStatistic(FeatureStatistic):
             - If 'lars_path', uses the lambda value where each feature/knockoff
             enters the lasso path (meaning becomes nonzero).
             This defaults to coef.
-        use_pyglm : bool
-            When fitting the group lasso, use the pyglm package if True (default).
-            Else, use the group_lasso package.
-        y_dist : str
-            One of "binomial" or "gaussian"
-        group_lasso : bool
-            If True, use a true group lasso. Else just use the sklearn ungrouped
-            lasso.
         antisym : str
             The antisymmetric function used to create (ungrouped) feature
             statistics. Three options: 
@@ -987,7 +850,13 @@ class LassoStatistic(FeatureStatistic):
             ``(2p, 2p)``-shaped precision matrix for the feature-knockoff 
             covariate distribution. This must be specified if ``debias=True``.
         kwargs : dict
-            Extra kwargs to pass to underlying Lasso classes
+            Extra kwargs to pass to underlying Lasso classes.
+
+        Notes
+        -----
+        When not using the group lasso, one can specify choice(s) of 
+        regularization parameter using the ``alphas`` keyword argument.
+        Otherwise, use the ``reg_vals`` keyword argument.
 
         Returns
         -------
@@ -1004,32 +873,25 @@ class LassoStatistic(FeatureStatistic):
             groups = np.arange(1, p + 1, 1)
 
         # Check if y_dist is gaussian, binomial, poisson
-        kwargs["y_dist"] = parse_y_dist(y)
+        kwargs["y_dist"] = kwargs.get("y_dist", parse_y_dist(y))
 
         # Step 1: Calculate Z statistics
         zstat = str(zstat).lower()
         if zstat == "coef":
 
             # Fit (possibly group) lasso
-            gl, inds, rev_inds = fit_group_lasso(
+            gl, inds, rev_inds = fit_lasso(
                 X=X,
                 Xk=Xk,
                 y=y,
-                groups=groups,
-                use_pyglm=use_pyglm,
-                group_lasso=group_lasso,
                 **kwargs,
             )
 
-            # Parse the expected output format based on which
-            # lasso package we are using
-            reg_lasso_flag = use_reg_lasso(groups) or (not group_lasso)
+            # Parse the expected output format 
             logistic_flag = parse_logistic_flag(kwargs)
 
             # Retrieve Z statistics
-            if use_pyglm and not reg_lasso_flag:
-                Z = gl.beta_[rev_inds]
-            elif reg_lasso_flag and logistic_flag:
+            if logistic_flag:
                 if gl.coef_.shape[0] != 1:
                     raise ValueError(
                         "Unexpected shape for logistic lasso coefficients (sklearn)"
@@ -1295,7 +1157,7 @@ class RandomForestStatistic(FeatureStatistic):
 
         # By default, all variables are their own group
         if groups is None:
-            groups = np.arange(0, p, 1)
+            groups = np.arange(1, p+1)
         self.groups = groups
 
         # Parse y_dist, initialize model
@@ -1308,7 +1170,8 @@ class RandomForestStatistic(FeatureStatistic):
         else:
             self.model = ensemble.RandomForestClassifier(**kwargs)
 
-        # Fit model, get Z statistics
+        # Fit model, get Z statistics.
+        # Note this does the randomization of features by itself.
         self.model.fit(features, y)
         feature_importance = str(feature_importance).lower()
         if feature_importance == "default":
@@ -1390,10 +1253,9 @@ class DeepPinkStatistic(FeatureStatistic):
             "sum" and "avg".
         cv_score : bool
             If true, score the feature statistic's predictive accuracy
-            using cross validation. This is extremely expensive for random
-            forests.
+            using cross validation. This is extremely expensive for deeppink.
         kwargs : dict
-            Extra kwargs to pass to underlying RandomForest class
+            Extra kwargs to pass to underlying deeppink class (in kpytorch)
 
         Returns
         -------
@@ -1412,22 +1274,22 @@ class DeepPinkStatistic(FeatureStatistic):
         p = X.shape[1]
         features = np.concatenate([X, Xk], axis=1)
 
-        # Randomize coordinates to make sure everything is symmetric
-        self.inds = np.arange(0, 2 * p, 1)
-        self.rev_inds = np.arange(0, 2 * p, 1)
+        # The deeppink model class will shuffle statistics,
+        # but for compatability we create indices anyway
+        self.inds = np.arange(2*p)
+        self.rev_inds = self.inds
 
         # By default, all variables are their own group
         if groups is None:
-            groups = np.arange(0, p, 1)
+            groups = np.arange(1, p+1)
         self.groups = groups
 
         # Parse y_dist, hidden_sizes, initialize model
         y_dist = parse_y_dist(y)
         if "hidden_sizes" not in kwargs:
             kwargs["hidden_sizes"] = [min(n, p)]
-        self.model = deeppink.DeepPinkModel(
-            p=p, inds=self.inds, rev_inds=self.inds, **kwargs
-        )
+        self.model = deeppink.DeepPinkModel(p=p, **kwargs)
+
         # Train model
         self.model.train()
         self.model = deeppink.train_deeppink(self.model, features, y, **train_kwargs)
@@ -1436,7 +1298,7 @@ class DeepPinkStatistic(FeatureStatistic):
         # Get Z statistics
         feature_importance = str(feature_importance).lower()
         if feature_importance == "deeppink":
-            self.Z = self.model.feature_importances()
+            self.Z = self.model.feature_importances(weight_scores=True)
         elif feature_importance == "unweighted":
             self.Z = self.model.feature_importances(weight_scores=False)
         elif feature_importance == "swap":
