@@ -245,10 +245,10 @@ class MetropolizedKnockoffSampler(KnockoffSampler):
             # This is more numerically stable for super sparse Q
             Q = np.linalg.inv(V)
         if undir_graph is not None:
-            warnings.filterwarnings("ignore")
-            mask = nx.to_numpy_matrix(undir_graph)
-            warnings.resetwarnings()
-            np.fill_diagonal(mask, 1)
+            with warnings.catch_warnings():
+                warnings.simplefilter('ignore')
+                mask = nx.to_numpy_matrix(undir_graph)
+                np.fill_diagonal(mask, 1)
             # Handle case where the graph is entirely dense
             if (mask == 0).sum() > 0 and not cov_est:
                 max_nonedge = np.max(np.abs(Q[mask == 0]))
@@ -864,7 +864,9 @@ class MetropolizedKnockoffSampler(KnockoffSampler):
         # Fast_exp function is helpful for profiling
         # (to see how much time is spent here)
         def fast_exp(ld_ratio, lq_ratio, Fj_ratio):
-            return np.exp((ld_ratio + lq_ratio + Fj_ratio).astype(np.float32))
+            return np.exp((
+                np.minimum(ld_ratio + lq_ratio + Fj_ratio, constants.MAXEXP32)
+            ).astype(np.float32))
 
         acc_prob = self.gamma * np.minimum(1, fast_exp(ld_ratio, lq_ratio, Fj_ratio))
 
@@ -987,7 +989,7 @@ class MetropolizedKnockoffSampler(KnockoffSampler):
                 self.affected_vars[j] += [j2]
 
         # Store pattern of TRUE acceptances / rejections
-        self.acceptances = np.zeros((self.n, self.p)).astype(np.bool)
+        self.acceptances = np.zeros((self.n, self.p)).astype(bool)
         self.final_acc_probs = np.zeros((self.n, self.p))
 
         # Proposals
@@ -1034,7 +1036,7 @@ class MetropolizedKnockoffSampler(KnockoffSampler):
             self.final_acc_probs[:, j] = acc_prob
 
             # Sample to get actual acceptances
-            self.acceptances[:, j] = np.random.binomial(1, acc_prob).astype(np.bool)
+            self.acceptances[:, j] = np.random.binomial(1, acc_prob).astype(bool)
 
             # Store knockoffs
             mask = self.acceptances[:, j] == 1
