@@ -8,6 +8,7 @@ from . import constants
 
 # These methods don't require approximation
 NON_APPROX_METHODS = ["equicorrelated", "eq", "ci", "ciknock"]
+DEFAULT_SOLVERS = {"mvr":"cd", "maxent":"cd", "sdp":"sdp"}
 
 def parse_method(method, groups, p=None):
     """ Decides which method to use to create the knockoff S matrix """
@@ -21,13 +22,12 @@ def parse_method(method, groups, p=None):
         return "mvr"
     else:
         return "sdp"
-    return method
-
 
 def divide_computation(Sigma, max_block):
     """
-    Approximates Sigma as a block-diagonal matrix.
-    
+    Approximates a correlation matrix Sigma as a block-diagonal matrix
+    using hierarchical clustering. Roughly follows the R knockoff package.
+
     Parameters
     ----------
     Sigma : np.ndarray
@@ -39,14 +39,7 @@ def divide_computation(Sigma, max_block):
     -------
     blocks : np.ndarray
         ``(p, )``-shaped numpy array where ``blocks[i] == j`` indicates
-        that variable ``i`` belongs to block ``j``. 
-    """
-
-
-def divide_computation(Sigma, max_block):
-    """
-    Approximates a correlation matrix Sigma as a block-diagonal matrix
-    using hierarchical clustering. Roughly follows the R knockoff package.
+        that variable ``i`` belongs to block ``j``.
     """
 
     # Correlation tree. We add noise to deal with highly structured Sigma.
@@ -159,7 +152,7 @@ def compute_smatrix_factored(
 
     # Compute factor model if not inputted
     if D is None or U is None:
-        raise NotImplementedError("TODO")
+        D, U = utilities.estimate_factor(Sigma, num_factors=num_factors)
     p = D.shape[0]
     k = U.shape[1]
     
@@ -168,6 +161,8 @@ def compute_smatrix_factored(
         S = mrc.solve_mvr_factored(D=D, U=U, **kwargs)
     elif method == 'maxent' or method == 'mmi':
         S = mrc.solve_maxent_factored(D=D, U=U, **kwargs)
+    elif method == 'sdp' or method == 'asdp':
+        S = mrc._solve_maxent_sdp_factored(D=D, U=U, solve_sdp=True, **kwargs)
     else: 
         raise NotImplementedError(f"Factor model solver is not implemented for method={method}")
 
@@ -183,11 +178,12 @@ def compute_smatrix(
     Sigma,
     groups=None,
     method=None,
-    solver="cd",
+    solver=None,
     how_approx='blockdiag',
     max_block=1000,
     num_factors=20,
     num_processes=1,
+    line_search=True,
     D=None,
     U=None,
     **kwargs,
@@ -209,9 +205,9 @@ def compute_smatrix(
     method : str
         Method for constructing S-matrix. One of mvr, maxent, mmi, sdp, equicorrelated, ci.
     solver : str
-        Method for solving mrc knockoffs. One of 'cd' (coordinate descent) 
-        or 'psgd' (projected gradient descent). Coordinate descent is 
-        highly recommended.
+        Method for solving for knockoffs. One of 'cd' (coordinate descent),
+        'sdp' (default sdp solver), or 'psgd' (projected gradient descent). 
+        Coordinate descent is recommended for MRC knockoffs.
     how_approx : str
         How to approximate the covariance matrix to speed up computation. 
         - If 'blockdiag', approximates Sigma as a block-diagonal matrix.
@@ -230,6 +226,8 @@ def compute_smatrix(
     U : np.ndarray
         ``(p, k)``-shaped matrix for factor model. Usually k << p. Only used 
         if how_approx='factor'. This is optional if Sigma is not None.
+    line_search : bool
+        If false, mrc methods do not do a line search after blockdiag approximation.
     kwargs : dict
         kwargs to pass to one of the wrapped S-matrix solvers.
 
@@ -256,6 +254,15 @@ def compute_smatrix(
     if method is not None:
         method = str(method).lower()
     method = parse_method(method, groups, None)
+
+    # Parse solver
+    if solver is None:
+        if method in DEFAULT_SOLVERS:
+            solver = DEFAULT_SOLVERS[method]
+        else:
+            solver = "cd"
+    solver = str(solver).lower()
+
     # These two are the same
     if method == 'mmi':
         method = 'maxent'
@@ -328,13 +335,13 @@ def compute_smatrix(
             num_iter=kwargs.get("num_iter", 10),
         )
         # Line search for MRC methods
-        smoothing = kwargs.get("smoothing", 0)
-        if method == "mvr":
-            loss_fn = mrc.mvr_loss
-        elif method == "maxent":
-            loss_fn = mrc.maxent_loss
         best_gamma = 1
-        if method in ["mvr", "maxent"]:
+        if line_search and method in ["mvr", "maxent"]:
+            smoothing = kwargs.get("smoothing", 0)
+            if method == "mvr":
+                loss_fn = mrc.mvr_loss
+            elif method == "maxent":
+                loss_fn = mrc.maxent_loss
             best_loss = loss_fn(Sigma=Sigma, S=S, smoothing=smoothing)
             for gamma in np.arange(20)/10:
                 loss = loss_fn(Sigma=Sigma, S=gamma * S, smoothing=smoothing)
@@ -344,22 +351,31 @@ def compute_smatrix(
 
         return S * best_gamma * scale_matrix
 
-    # Currently cd solvers cannot handle group knockoffs
+    # Currently cd maxent solver cannot handle group knockoffs
     # (this is todo)
-    if not np.all(groups == np.arange(1, p + 1, 1)):
+    no_grouping = np.all(groups == np.arange(1, p + 1, 1))
+    if not no_grouping and method == 'maxent':
         solver = "psgd"
     if (method == "mvr" or method == "maxent") and solver == "psgd":
         # Check for imports
         utilities.check_kpytorch_available(purpose="group MRC knockoffs OR PSGD solver")
         from .kpytorch import mrcgrad
-
         S = mrcgrad.solve_mrc_psgd(Sigma=Sigma, groups=groups, method=method, **kwargs)
     elif method == "mvr":
-        S = mrc.solve_mvr(Sigma=Sigma, **kwargs)
+        S = mrc.solve_mvr(Sigma=Sigma, groups=groups, **kwargs)
     elif method == "maxent":
         S = mrc.solve_maxent(Sigma=Sigma, **kwargs)
     elif method == "sdp" or method == "asdp":
-        S = mac.solve_group_SDP(Sigma, groups, **kwargs,)
+        # Currently cd sdp solver cannot handle group knockoffs
+        # (this is todo)
+        if solver == "sdp" or not no_grouping:
+            S = mac.solve_group_SDP(Sigma, groups, **kwargs,)
+        elif solver == 'cd' and no_grouping:
+            S = mrc._solve_maxent_sdp_cd(
+                Sigma=Sigma, solve_sdp=True, **kwargs
+            )
+        else:
+            raise ValueError(f"solver={solver} is not supported for {method} knockoffs")
     elif method == "ciknock" or method == "ci":
         S = mrc.solve_ciknock(Sigma, **kwargs)
     elif method == "equicorrelated" or method == "eq":
