@@ -2,8 +2,12 @@ import numpy as np
 import scipy as sp
 import unittest
 import pytest
-from .context import knockpy
-
+# for regular pytest calls
+try:
+    from .context import knockpy
+# for running directly with python
+except ImportError:
+    from context import knockpy
 
 from knockpy import dgp
 
@@ -145,7 +149,7 @@ class TestSampleData(unittest.TestCase):
         _, _, beta, _, _ = dgprocess.sample_data(
             p=p, sparsity=1, coeff_size=1, coeff_dist="normal", sign_prob=0
         )
-        expected = 1
+        expected = 0
         mean_est = beta.mean()
         self.assertTrue(
             np.abs(mean_est - expected) < 0.1,
@@ -409,6 +413,18 @@ class TestSampleData(unittest.TestCase):
             err_msg=f"random AR1 gen has unexpected avg rho {mean_rho} vs {expected} ",
         )
 
+        # Check that maxcorr works
+        max_corr = 0.8
+        dgprocess = dgp.DGP()
+        _, _, _, _, Sigma = dgprocess.sample_data(
+            p=500, method="AR1", a=10, b=1, max_corr=max_corr
+        )
+        obs_max_corr = np.max(np.diag(Sigma, 1))
+        self.assertTrue(
+            max_corr >= obs_max_corr - 1e-5, # float errors
+            f"observed max corr {obs_max_corr} > max_corr {max_corr}"
+        )
+
     def test_nested_AR1(self):
 
         # Check that a, b parameters work
@@ -453,13 +469,12 @@ class TestSampleData(unittest.TestCase):
             err_msg=f"random unifdot generation {Sigma} unexpectedly deviates from the {expected}",
         )
 
-    @pytest.mark.slow
     def test_dirichlet_matrices(self):
         """ Simple test that ensures there are no errors, we get corr matrix 
 		with expected eigenvalues"""
 
         # Try one with low temp
-        p = 2000
+        p = 300
         temp = 0.1
         np.random.seed(110)
         dgprocess = dgp.DGP()
@@ -520,7 +535,6 @@ class TestSampleData(unittest.TestCase):
         np.random.seed(110)
         dgprocess = dgp.DGP()
         _, _, _, Q, V = dgprocess.sample_data(p=p, delta=delta, method="ver")
-
         prop_nonzero = (np.abs(V) > 0.001).mean()
         self.assertTrue(
             abs(prop_nonzero - delta) < 0.02,
@@ -533,6 +547,23 @@ class TestSampleData(unittest.TestCase):
             "True (V)ErdosRenyi sampler fails to give correct mean val",
         )
 
+        # Try er = V with maxcorr
+        max_corr = 0.1
+        delta = 0.05
+        dgprocess = dgp.DGP()
+        _, _, _, _, V = dgprocess.sample_data(p=p, delta=delta, method="ver", max_corr=max_corr)
+        np.testing.assert_array_almost_equal(
+            np.diag(V),
+            np.ones(p),
+            err_msg=f"After setting max_corr={max_corr}, ER cov is not corr. matrix")
+        hV = V - np.diag(np.diag(V)) # zero out diagonals 
+        obs_max_corr = np.abs(hV).max()
+        self.assertTrue(
+            obs_max_corr <= max_corr + 1e-5,
+            f"For VER, obs_max_corr {obs_max_corr} >= max_corr {max_corr}"
+        )
+
+
     def test_tblock_sample(self):
 
         # Fake data --> we want the right cov matrix
@@ -591,8 +622,8 @@ class TestSampleData(unittest.TestCase):
         # Check that we get a decent correlation matrix
         # with the right type of Q matrix
         np.random.seed(110)
-        n = 150000
-        p = 9
+        n = 60000
+        p = 4
         dgprocess = dgp.DGP()
         X,_,_,_,V = dgprocess.sample_data(n=n, p=p, method="ising", x_dist="gibbs",)
         gibbs_graph = dgprocess.gibbs_graph
@@ -642,6 +673,37 @@ class TestSampleData(unittest.TestCase):
 
         self.assertRaisesRegex(ValueError, "x_dist must be one of", bad_xdist)
 
+class TestGroupings(unittest.TestCase):
+
+    def test_group_creation(self):
+
+        # Random covariance matrix
+        np.random.seed(110)
+        dgprocess = knockpy.dgp.DGP()
+        dgprocess.sample_data(p=100)
+        # Check within-group correlations are less than 0.5
+        Sigma = dgprocess.Sigma
+        #helper_Sigma = dgprocess.Sigma.copy()
+        #helper_Sigma = helper_Sigma - 1 * np.diag(np.diag(helper_Sigma))
+        for cutoff in [0.1, 0.5, 0.9]:
+            groups = knockpy.dgp.create_grouping(Sigma, cutoff=cutoff, method='single')
+            for j in np.unique(groups):
+                between_group_corrs = Sigma[groups==j][:, groups!=j]
+                maxcorr = np.max(np.abs(between_group_corrs))
+                self.assertTrue(
+                     maxcorr <= cutoff,
+                    f"Max between-group corr is {maxcorr} > cutoff {cutoff}"
+                )
+
+        groups = knockpy.dgp.create_grouping(Sigma, cutoff=1, method='single')
+        np.testing.assert_array_almost_equal(
+            np.sort(groups),
+            np.arange(1, Sigma.shape[0] + 1),
+            err_msg=f"When cutoff=1, groups are not trivial groups"
+        )
 
 if __name__ == "__main__":
-    unittest.main()
+    import pytest
+    import sys
+    pytest.main(sys.argv)
+    #unittest.main()
