@@ -6,8 +6,7 @@ import unittest
 from .context import knockpy
 from .context import file_directory
 
-from knockpy import utilities
-from knockpy import dgp
+from knockpy import dgp, mlr, utilities
 from knockpy.knockoff_filter import KnockoffFilter
 
 NUM_REPS = 1
@@ -39,7 +38,7 @@ class TestFdrControl(unittest.TestCase):
         S=None,
         infer_sigma=False,
         test_grouped=True,
-        S_method="sdp",
+        S_method="mvr",
         **kwargs,
     ):
 
@@ -75,13 +74,15 @@ class TestFdrControl(unittest.TestCase):
         init_filter_kwargs = {}
         init_filter_kwargs["ksampler"] = filter_kwargs.pop("ksampler", "gaussian")
         init_filter_kwargs["fstat"] = filter_kwargs.pop("fstat", "lasso")
+        knockoff_kwargs = filter_kwargs.pop('knockoff_kwargs', {})
+
 
         for name, groups in zip([name1, name2], [groups1, groups2]):
 
             if not test_grouped and np.all(groups==groups2):
                 continue
 
-            # Solve SDP
+            # Solve for S matrix
             if S is None and not fixedX and not infer_sigma:
                 ksampler = knockpy.knockoffs.GaussianSampler(
                     X=X0, Sigma=Sigma, groups=groups, method=S_method,
@@ -120,11 +121,10 @@ class TestFdrControl(unittest.TestCase):
                 knockoff_filter = KnockoffFilter(**init_filter_kwargs)
 
                 # Knockoff kwargs
-                knockoff_kwargs = {
-                    "S": S,
-                    "invSigma": invSigma_arg,
-                    "verbose": False,
-                }
+                knockoff_kwargs['S'] = S
+                knockoff_kwargs['invSigma'] = invSigma_arg
+                knockoff_kwargs['verbose'] = False
+
                 if "df_t" in kwargs:
                     knockoff_kwargs["df_t"] = kwargs["df_t"]
                 if "x_dist" in kwargs:
@@ -142,7 +142,16 @@ class TestFdrControl(unittest.TestCase):
                     fdr=q,
                     **filter_kwargs,
                 )
-                del knockoff_filter
+
+                # Check null W-statistics are symmetric
+                pos_prop = (knockoff_filter.W[group_nonnulls == 0] > 0).mean() 
+                pos_prop_se = np.sqrt(pos_prop * (1 - pos_prop) / (1 - group_nonnulls).sum())
+                Zstat = (pos_prop - 0.5) / pos_prop_se
+                pval = 1 - stats.norm.cdf(Zstat)
+                self.assertTrue(
+                    pval >= 0.001,
+                    msg=f"MX filter null W-stats have pos_prob {pos_prop} with p={p} and pval={pval}",
+                )
 
                 # Calculate fdp
                 fdp = np.sum(selections * (1 - group_nonnulls)) / max(
@@ -150,6 +159,8 @@ class TestFdrControl(unittest.TestCase):
                 )
                 fdps.append(fdp)
 
+                del knockoff_filter
+
             fdps = np.array(fdps)
             fdr = fdps.mean()
             # fdr_se = fdps.std() / np.sqrt(reps)
@@ -347,6 +358,17 @@ class TestKnockoffFilter(TestFdrControl):
             filter_kwargs={"ksampler": "fx"},
         )
 
+        # Scenario 1: AR1, recycle, lsm
+        self.check_fdr_control(
+            reps=NUM_REPS,
+            n=150,
+            p=50,
+            method="AR1",
+            sparsity=0.1,
+            y_dist="gaussian",
+            filter_kwargs={"ksampler": "fx", "fstat": "lsm"},
+        )
+
     @pytest.mark.slow
     def test_deeppink_control(self):
         if not TORCH_AVAILABLE:
@@ -447,6 +469,50 @@ class TestKnockoffFilter(TestFdrControl):
             filter_kwargs={"fstat_kwargs": {"use_lars": True}},
         )
 
+    @pytest.mark.quick
+    def test_mlr_fit(self):
+        """Tests that MLR statistics work"""
+        n = 200
+        p = 50
+        # 1. FX
+        self.check_fdr_control(
+            reps=NUM_REPS,
+            n=n,
+            p=p,
+            y_dist='gaussian',
+            test_grouped=False,
+            filter_kwargs={
+                "ksampler":"fx",
+                "fstat":"mlr",
+                "fstat_kwargs":{"n_iter":10, "chains":2, "num_mixture":4},
+            },
+        )
+        # 2. MX
+        self.check_fdr_control(
+            reps=NUM_REPS,
+            n=n,
+            p=p,
+            y_dist='gaussian',
+            test_grouped=True,
+            filter_kwargs={
+                "fstat":"mlr",
+                "fstat_kwargs":{"n_iter":10, "chains":2},
+            },        
+        )
+
+        # 3. splines
+        self.check_fdr_control(
+            reps=NUM_REPS,
+            n=n,
+            p=p,
+            y_dist='gaussian',
+            test_grouped=True,
+            filter_kwargs={
+                "fstat":"mlr_spline",
+                "fstat_kwargs":{"n_iter":10, "chains":2, "degree":4, "n_knots":3},
+            },
+        )
+
     @pytest.mark.slow
     def test_factor_model(self):
 
@@ -469,6 +535,26 @@ class TestKnockoffFilter(TestFdrControl):
         time_factored = time.time() - time0
         print(f"Factored time is {time_factored}")
 
+        time0 = time.time()
+        self.check_fdr_control(
+            n=n,
+            p=p,
+            method='blockequi',
+            gamma=gamma,
+            rho=rho,
+            infer_sigma=True,
+            S_method="mvr",
+            filter_kwargs={
+                "knockoff_kwargs":{
+                    'how_approx':'factor'
+                }, 
+                'fstat':'margcorr'
+            },
+            test_grouped=False,
+        )
+        time_apprx_factored = time.time() - time0
+        print(f"Apprx_factored time is {time_apprx_factored}")
+
         time0 = time.time()
         self.check_fdr_control(
             n=n,
@@ -485,9 +571,12 @@ class TestKnockoffFilter(TestFdrControl):
         print(f"Unfactored time is {time_unfactored}")
         self.assertTrue(
             1.5*time_factored < time_unfactored,
-            msg=f"time for factor apprx ({time_factored}) > 1.5*time for no apprx ({time_unfactored})"
+            msg=f"time for factor assumption ({time_factored}) > 1.5*time for no apprx ({time_unfactored})"
+        )
+        self.assertTrue(
+            1.2*time_apprx_factored < time_unfactored,
+            msg=f"time for factor apprx ({time_apprx_factored}) > 1.2*time for no apprx ({time_unfactored})"
         )
-
 
 
     @pytest.mark.quick
