@@ -2,8 +2,12 @@ import numpy as np
 import unittest
 import sklearn.naive_bayes
 import sklearn.neural_network
-from .context import knockpy
-
+# for regular pytest calls
+try:
+    from .context import knockpy
+# for running directly with python
+except ImportError:
+    from context import knockpy
 from knockpy import knockoff_stats as kstats
 from knockpy import utilities, dgp
 from knockpy.knockoff_stats import data_dependent_threshhold
@@ -82,7 +86,7 @@ class KStatVal(unittest.TestCase):
             X, Xk, y, groups=groups, **fstat_kwargs,
         )
         W = fstat.W
-        T = data_dependent_threshhold(W, fdr=0.2)
+        T = data_dependent_threshhold(W, fdr=0.2)        
 
         # Test L2 norm
         m = np.unique(groups).shape[0]
@@ -108,6 +112,9 @@ class KStatVal(unittest.TestCase):
             msg=f"Power {power} for {fstat_name} in equicor case (n={n},p={p},rho={rho}, y_dist {y_dist}, grouped={group_features}) should be > {min_power}. W stats are {W}, beta is {beta}",
         )
 
+        # Test symmetric property of null W
+        
+
 
 class TestFeatureStatistics(KStatVal):
     """ Tests fitting of ols, lasso, ridge, margcorr, random forest """
@@ -210,82 +217,6 @@ class TestFeatureStatistics(KStatVal):
             min_power=0.8,
         )
 
-    def test_pyglm_group_lasso_fit(self):
-
-        pyglm_kwargs = {
-            "use_pyglm": True,
-            "max_iter": 20,
-            "tol": 5e-2,
-            "learning_rate": 3,
-            "group_lasso": True,
-        }
-        self.check_kstat_fit(
-            fstat=kstats.LassoStatistic(),
-            fstat_name="Pyglm solver",
-            fstat_kwargs=pyglm_kwargs,
-            n=500,
-            p=200,
-            rho=0.2,
-            coeff_size=5,
-            sparsity=0.5,
-            seed=110,
-            min_power=0.5,
-            group_features=True,
-            max_l2norm=np.inf,
-        )
-
-        # Repeat for logistic case
-        self.check_kstat_fit(
-            fstat=kstats.LassoStatistic(),
-            fstat_name="Pyglm solver",
-            fstat_kwargs=pyglm_kwargs,
-            n=500,
-            p=100,
-            rho=0.2,
-            coeff_size=5,
-            sparsity=0.5,
-            seed=110,
-            min_power=0.5,
-            group_features=True,
-            y_dist="binomial",
-            max_l2norm=np.inf,
-        )
-
-    def test_vanilla_group_lasso_fit(self):
-
-        glasso_kwargs = {"use_pyglm": False, "group_lasso": True}
-        self.check_kstat_fit(
-            fstat=kstats.LassoStatistic(),
-            fstat_name="Vanilla group lasso solver",
-            fstat_kwargs=glasso_kwargs,
-            n=500,
-            p=200,
-            rho=0.2,
-            coeff_size=5,
-            sparsity=0.5,
-            seed=110,
-            min_power=0.5,
-            group_features=True,
-            max_l2norm=np.inf,
-        )
-
-        # Repeat for logistic case
-        self.check_kstat_fit(
-            fstat=kstats.LassoStatistic(),
-            fstat_name="Vanilla group lasso solver",
-            fstat_kwargs=glasso_kwargs,
-            n=300,
-            p=100,
-            rho=0.2,
-            coeff_size=5,
-            sparsity=0.5,
-            seed=110,
-            min_power=0.5,
-            group_features=True,
-            y_dist="binomial",
-            max_l2norm=np.inf,
-        )
-
     def test_lasso_fit(self):
 
         # Lasso fit for Gaussian data
@@ -333,6 +264,24 @@ class TestFeatureStatistics(KStatVal):
             max_l2norm=np.inf,
         )
 
+        # Test the regularization parameters option
+        alpha = 50
+        dgprocess = knockpy.dgp.DGP()
+        dgprocess.sample_data(n=300, p=100, sparsity=0.1)
+        kfilter = knockpy.knockoff_filter.KnockoffFilter(
+            fstat='lasso', ksampler='gaussian'
+        )
+        kfilter.forward(
+            X=dgprocess.X,
+            y=dgprocess.y,
+            Sigma=dgprocess.Sigma,
+            fstat_kwargs={"alphas":alpha}
+        )
+        self.assertTrue(
+            np.all(kfilter.W == 0),
+            f"lasso with alpha={alpha} should zero out coefs, instead W={kfilter.W}"
+        )
+
     def test_antisymmetric_fns(self):
 
         n = 100
@@ -426,29 +375,6 @@ class TestFeatureStatistics(KStatVal):
             msg=f"cv scoring fails for ols_stat as cv_score={ols_stat.score} >= 2",
         )
 
-        # 3. Test that throws correct error for non-sklearn backend
-        def non_sklearn_backend_cvscore():
-            dgprocess = dgp.DGP()
-            X, y, beta, _, corr_matrix = dgprocess.sample_data(
-                n=n, p=p, y_dist="binomial", coeff_size=100, sign_prob=1
-            )
-            groups = np.random.randint(1, p + 1, size=(p,))
-            group = utilities.preprocess_groups(groups)
-            pyglm_logit = kstats.LassoStatistic()
-            pyglm_logit.fit(
-                X,
-                knockoffs,
-                y,
-                use_pyglm=True,
-                group_lasso=True,
-                groups=groups,
-                cv_score=True,
-            )
-
-        self.assertRaisesRegex(
-            ValueError, "must be sklearn estimator", non_sklearn_backend_cvscore
-        )
-
     def test_debiased_lasso(self):
 
         # Create data generating process
@@ -526,7 +452,7 @@ class TestFeatureStatistics(KStatVal):
         self.check_kstat_fit(
             fstat=kstats.RidgeStatistic(),
             fstat_name="Sklearn ridge",
-            n=160,
+            n=200,
             p=100,
             rho=0.7,
             coeff_size=5,
@@ -540,7 +466,7 @@ class TestFeatureStatistics(KStatVal):
         self.check_kstat_fit(
             fstat=kstats.RidgeStatistic(),
             fstat_name="Sklearn ridge",
-            n=160,
+            n=200,
             p=100,
             rho=0.7,
             coeff_size=5,
@@ -555,7 +481,7 @@ class TestFeatureStatistics(KStatVal):
         self.check_kstat_fit(
             fstat=kstats.RidgeStatistic(),
             fstat_name="Sklearn ridge",
-            n=150,
+            n=200,
             p=100,
             rho=0.7,
             coeff_size=5,
@@ -871,6 +797,36 @@ class TestDataThreshhold(unittest.TestCase):
             err_msg=f"Incorrect data dependent threshhold (batched): Ts should be {expected}, not {Ts}",
         )
 
+class TestHelpers(unittest.TestCase):
+    """ tests miscallaneous helper functions """
+
+    def test_resid_variance_estimation(self):
+
+        # Create fake data and knockoffs in low and high-dimensional setting
+        np.random.seed(111)
+        p = 100
+        for n in [int(p / 2), int(2 * p) + 5, int(3 * p)]:      
+            dgprocess = knockpy.dgp.DGP()
+            dgprocess.sample_data(n=n, p=p, sparsity=0.1)
+            if n > 2 * p:
+                ksampler = knockpy.knockoffs.FXSampler(X=dgprocess.X)
+            else:
+                ksampler = knockpy.knockoffs.GaussianSampler(
+                    X=dgprocess.X, Sigma=dgprocess.Sigma
+                )
+            ksampler.sample_knockoffs()
+            # Compute residual variance
+            hat_sigma2 = kstats.compute_residual_variance(
+                dgprocess.X, ksampler.Xk, dgprocess.y
+            )
+            if n > p:
+                self.assertTrue(
+                    hat_sigma2 < 1.5 and hat_sigma2 > 0.66,
+                    f"Resid. var. est is poor: hat_sigma2={hat_sigma2} (target=1) for n={n}, p={p}"
+                )
 
 if __name__ == "__main__":
-    unittest.main()
+    import pytest
+    import sys
+    pytest.main(sys.argv)
+    #unittest.main()
