@@ -3,7 +3,13 @@ import warnings
 import numpy as np
 import scipy as sp
 import unittest
-from .context import knockpy
+
+# for regular pytest calls
+try:
+    from .context import knockpy
+# for running directly with python
+except ImportError:
+    from context import knockpy
 from knockpy import dgp, utilities, mac, mrc, smatrix, knockoffs
 
 try:
@@ -175,7 +181,7 @@ class TestSDP(CheckSMatrix):
             decimal=2,
             err_msg="solve_group_SDP does not produce optimal S matrix (blockequi graphs)",
         )
-        self.check_S_properties(corr_matrix, S_triv, trivial_groups)
+        self.check_S_properties(corr_matrix, S_triv, trivial_groups)        
 
         # Repeat for gaussian_knockoffs method
         ksampler = knockoffs.GaussianSampler(
@@ -227,27 +233,40 @@ class TestSDP(CheckSMatrix):
 
         # Test non-group SDP on equicorrelated cov matrix
         p = 100
-        rho = 0.8
-        V = rho * np.ones((p, p)) + (1 - rho) * np.eye(p)
-        S = mac.solve_group_SDP(V, verbose=True)
-        expected = (2 - 2 * rho) * np.eye(p)
-        np.testing.assert_almost_equal(
-            S,
-            expected,
-            decimal=2,
-            err_msg="solve_SDP does not produce optimal S matrix (equicorrelated graph)",
-        )
+        rhos = [0.6, 0.8, 0.9]
+        solvers = ['sdp', 'cd']
+        scales = [1, 5]
+        for scale in scales:
+            for rho in rhos:
+                V = scale * rho * np.ones((p, p)) + scale * (1 - rho) * np.eye(p)
+                for solver in solvers:
+                    S = smatrix.compute_smatrix(Sigma=V, method="sdp", solver=solver, verbose=True)
+                    expected = (2 - 2 * rho) * np.eye(p)
+                    np.testing.assert_almost_equal(
+                        S / scale,
+                        expected,
+                        decimal=2,
+                        err_msg=f"compute_smatrix produces incorrect S_SDP for equicorr, rho={rho}, scale={scale}, solver={solver}",
+                    )
 
-        # Repeat for scaled cov matrix
-        scale = 5
-        V = scale * V
-        S = smatrix.compute_smatrix(Sigma=V, method="sdp", verbose=True)
-        expected = (2 - 2 * rho) * np.eye(p)
+    def test_cd_SDP(self):
+        """
+        Test that CD / SDP solvers get similar answers in a hard problem
+        """
+
+        np.random.seed(110)
+        V = utilities.cov2corr(dgp.ErdosRenyi(p=100, tol=1e-1))
+        S_cd = smatrix.compute_smatrix(
+            Sigma=V, method="sdp", solver="cd", verbose=True, mu=0.95, num_iter=100
+        )
+        mac_cd = np.diag(S_cd).mean()
+        S_sdp = smatrix.compute_smatrix(Sigma=V, method="sdp", solver="sdp")
+        mac_sdp = np.diag(S_sdp).mean()
         np.testing.assert_almost_equal(
-            S / scale,
-            expected,
+            mac_sdp,
+            mac_cd,
             decimal=2,
-            err_msg="compute_smatrix does not produce optimal S matrix for scaled equicorr graph",
+            err_msg=f"compute_smatrix mac differs between SDP/CD solvers ({mac_sdp}, {mac_cd}, respectively"
         )
 
     def test_sdp_tolerance(self):
@@ -673,6 +692,39 @@ class TestMRCSolvers(CheckSMatrix):
                 msg=f"For {method}, coord descent mmi solver has higher loss {cd_mmi_loss} v. PSGD {psgd_mmi_loss}",
             )
 
+    def test_cd_group_mvr(self):
+
+        np.random.seed(110)
+        p = 100
+        for method in ['ar1', 'ver', 'qer']:
+            # Sample data
+            dgprocess = knockpy.dgp.DGP()
+            dgprocess.sample_data(method=method, p=p)
+            Sigma = dgprocess.Sigma
+
+            # Sanity check: grouped solver gets right soln for ungrouped
+            S1 = mrc._solve_mvr_grouped(
+                Sigma, groups=np.arange(1, p+1), converge_tol=1e-2, verbose=True
+            )
+            S2 = mrc.solve_mvr(Sigma)
+            Sdiff = S1 - S2
+            mean_diff = np.abs(Sdiff[Sdiff != 0]).mean()
+            err_msg = f"For no groups, method={method}, avg diff. between grouped/ungrouped solvers is {mean_diff} > 1e-2"
+            err_msg = f"grouped solver: {S1} \n ungrouped solver: {S2} \n" + err_msg
+            self.assertTrue(
+                mean_diff < 1e-2, err_msg
+            )
+
+            # Check that adding groups lowers loss
+            loss_ungrouped = mrc.mvr_loss(Sigma, S2)
+            groups = np.around(np.arange(1, p+1)/2)
+            Sgroup = mrc.solve_mvr(Sigma, groups)
+            loss_grouped = mrc.mvr_loss(Sigma, Sgroup)
+            self.assertTrue(
+                loss_grouped < loss_ungrouped,
+                f"Grouped loss ({loss_grouped}) > ungrouped loss ({loss_ungrouped}) for method={method}" 
+            )
+
     def test_complex_group_solns(self):
         """
         Check the solutions of the PSGD solver
@@ -840,7 +892,7 @@ class TestFactorApprx(CheckSMatrix):
         ks = [1, 10, 20, 50]
         p = 150
         for k in ks:
-            for method in ['mvr', 'maxent']:   
+            for method in ['mvr', 'maxent', 'sdp']:   
                 D = np.random.uniform(low=0.01, high=1, size=(p,))
                 U = np.random.randn(p, k) / np.sqrt(k)
                 # Rescale to correlation matrix
@@ -853,7 +905,7 @@ class TestFactorApprx(CheckSMatrix):
                     Sigma=None, D=D, U=U, method=method, how_approx='factor'
                 )
                 S2 = smatrix.compute_smatrix(
-                   Sigma=Sigma, method=method, tol=0
+                   Sigma=Sigma, method=method, solver='cd', tol=0
                 )
                 np.testing.assert_array_almost_equal(
                     S1, S2, decimal=2,
@@ -949,7 +1001,7 @@ class TestKnockoffGen(CheckValidKnockoffs):
             ksampler = knockoffs.GaussianSampler(
                 X=X, Sigma=corr_matrix, S=S_bad, verbose=False
             )
-            ksampler.sample_knockoffs()
+            ksampler.sample_knockoffs(check_psd=True)
 
         self.assertRaisesRegex(
             np.linalg.LinAlgError,
@@ -1109,4 +1161,7 @@ class TestKnockoffGen(CheckValidKnockoffs):
 
 
 if __name__ == "__main__":
-    unittest.main()
+    import pytest
+    import sys
+    pytest.main(sys.argv)
+    #unittest.main()
\ No newline at end of file
